/*-------------------------------------------------------------------*
|GAIA - Software platform to integrate Speech Translation Technologies
|Funded by the European Union (LC-STAR project, IST-2001-32216) and
|the Spanish Government (ALIADO project, TIC2002-04447-C02).
|
|Copyright (C) Javier Perez		javier.perez@tsc.upc.edu
|	       Antonio Bonafonte	antonio.bonafonte@upc.edu
|	       Universitat Politecnica de Catalunya, Barcelona, Spain.
|
|Gaia (main program and related libraries and utilities) is free
|software; you can redistribute it and/or modify it under the terms of
|the GNU General Public License as published by the Free Software
|Foundation; either version 2 of the License, or (at your option) any
|later version.
|
|This program is distributed in the hope that it will be useful,
|but WITHOUT ANY WARRANTY; without even the implied warranty of
|MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|GNU General Public License for more details.
|
|You should have received a copy of the GNU General Public License
|along with this program; if not, write to the Free Software
|Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
|USA.
*--------------------------------------------------------------------*/
#include "includes/Sock.h"

//#include <pthread.h>
#include <signal.h>
#ifndef WIN32
#include <errno.h>
#else
#include <process.h>
#endif
#include <string>
#include <iostream>

#ifdef USE_PTHREAD
#include <pthread.h>
#endif

#include <string.h>

using namespace std;

#define LittleEndian 1
#define BigEndian    2
static int TestEndian();
static int fswap(void *_buffer, size_t size, size_t count);


#ifdef USE_PTHREAD
#define T_RETTHR void *
#else
#define T_RETTHR void
#endif


Sock::Sock() {
    //FIXME!!!
    //    swap = (TestEndian() == BigEndian);
      swap = (TestEndian() == LittleEndian);
#ifdef WIN32
    
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
 
    wVersionRequested = MAKEWORD( 2, 2 );
 
    err = WSAStartup( wVersionRequested, &wsaData );
    if ( err != 0 ) {
	/* Tell the user that we could not find a usable */
	/* WinSock DLL.                                  */
	cerr << "Wrong WinSock version!" << endl;
	return;
    }
 
    /* Confirm that the WinSock DLL supports 2.2.*/
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */
 
    if ( LOBYTE( wsaData.wVersion ) != 2 ||
	 HIBYTE( wsaData.wVersion ) != 2 ) {
	/* Tell the user that we could not find a usable */
	/* WinSock DLL.                                  */
	WSACleanup( );
	return; 
    }
 
    /* The WinSock DLL is acceptable. Proceed. */
#endif
}

Sock::Sock(int sock) {
    //FIXME!!
    //      swap = (TestEndian() == BigEndian);
      swap = (TestEndian() == LittleEndian);
  
    this->sock = sock;
#ifdef WIN32
    
    WORD wVersionRequested;
    WSADATA wsaData;
    int err;
 
    wVersionRequested = MAKEWORD( 2, 2 );
 
    err = WSAStartup( wVersionRequested, &wsaData );
    if ( err != 0 ) {
	/* Tell the user that we could not find a usable */
	/* WinSock DLL.                                  */
	cerr << "Wrong WinSock version!" << endl;
	return;
    }
 
    /* Confirm that the WinSock DLL supports 2.2.*/
    /* Note that if the DLL supports versions greater    */
    /* than 2.2 in addition to 2.2, it will still return */
    /* 2.2 in wVersion since that is the version we      */
    /* requested.                                        */
 
    if ( LOBYTE( wsaData.wVersion ) != 2 ||
	 HIBYTE( wsaData.wVersion ) != 2 ) {
	/* Tell the user that we could not find a usable */
	/* WinSock DLL.                                  */
	WSACleanup( );
	return; 
    }
 
    /* The WinSock DLL is acceptable. Proceed. */
#endif

}
int Sock::connect(string server, int port) {
    return connect(server.c_str(), port);
}
int Sock::connect(const char *server, int port) {

    struct sockaddr_in serv_adr;
    memset(&serv_adr, 0, sizeof(serv_adr));
    serv_adr.sin_family = AF_INET;
    serv_adr.sin_port = htons(port);

    struct hostent *host;
    if ((host = gethostbyname(server)) != 0)
	memcpy(&serv_adr.sin_addr, host->h_addr, host->h_length);
    else { 
//  	int i = h_errno;
	serv_adr.sin_addr.s_addr = inet_addr(server);
	if (serv_adr.sin_addr.s_addr == INADDR_NONE)
	    return 1;
    }

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return 2;

    if (::connect(sock, (struct sockaddr *) &serv_adr, sizeof(serv_adr)) < 0)
	return 3;

    return 0;
}

int Sock::listen(int port) {

    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	return 1;

    struct sockaddr_in serv_adr;
    memset(&serv_adr, 0, sizeof(serv_adr));
    serv_adr.sin_family = AF_INET;
    serv_adr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_adr.sin_port = htons(port);

    if (bind(sock, (struct sockaddr *) &serv_adr, sizeof(serv_adr)) < 0) {
	close();
	return 2;
    }

    if (::listen(sock, 5) < 0) {
	close();
	return 3;
    }

    return 0;
}

long Sock::read(void *buf, size_t count) const {
    return read(buf, sizeof(char), count);
}

long Sock::write(const string &data) const {
    unsigned long ret = write((const void *)data.c_str(), data.size()+1);
    if (ret != data.size()+1)
	return -1;
    else
	return 0; 
}

long Sock::read(string &data, size_t count) const {
    data.resize(count);
    return read((void *)data.c_str(), sizeof(char), count);
}

long Sock::write(const void *buf, size_t count) const {
    long ret = -1;
#ifndef WIN32
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	return ret;
#endif

    ret = ::send(sock, (char *)buf, (long)count, 0);

    return (ret); 
}

long Sock::write(const void *_buf, size_t size, size_t count) const {
    long ret = -1;
#ifndef WIN32
    if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	return ret;
#endif
    void *buf = (void *) _buf; //swap + swap => nothing .. the caller can think we don't touch
    if (swap && size > 1) fswap(buf, size, count);

    ret = ::send(sock, (char *)buf, (long)count * size, 0);

    if (swap && size > 1) fswap(buf, size, count);
    return (ret/size); //should be 'count'
}

long Sock::read(void *buf, size_t size, size_t count) const {
    long ret, nread = count * size;
    char *p = (char *)buf;

    while (nread) {
//  #ifndef WIN32
//  	if ((ret = ::read(sock, p, nread)) < 0)
//  #else
//  	    if ((ret = ::recv(sock, p, nread,0)) < 0)
//  #endif
	if((ret = ::recv(sock, p, nread, 0)) <= 0)
		return ret;
	nread -= ret;
	p += ret;
    }
    if (swap && size > 1) fswap(buf, size, count);
    return count;
}


struct fthread_arg {
    Sock *sock;
    void *arg;
    int (*f)(Sock *sock, void *arg);

    fthread_arg(Sock *_sock, void *_arg, int (*_f)(Sock *sock, void *arg)) {
	sock = _sock; arg = _arg; f = _f;
    }
};


T_RETTHR  fthread(void *tharg) {
    ((fthread_arg *)tharg)->f(((fthread_arg *)tharg)->sock, ((fthread_arg *)tharg)->arg);
    if (((fthread_arg *)tharg)->sock != NULL)
	delete ((fthread_arg *)tharg)->sock;
#ifdef USE_PTHREAD
    return 0;
#else
    return;
#endif
}

int Sock::select(long timeout_usec){
    fd_set set;
    struct timeval timeout;
    struct timeval* _ptimeout;

    FD_ZERO(&set);
    FD_SET(sock, &set);

    if (timeout_usec == -1)
	_ptimeout = NULL;
    else {
	timeout.tv_sec = 0;
	timeout.tv_usec = timeout_usec;
	_ptimeout = &timeout;
    }

    long int ret;
    do{
	ret = ::select (FD_SETSIZE, 
			&set, NULL, NULL, 
			_ptimeout);
    }
#ifndef WIN32 
    while( ret==-1L && *(__errno_location())==EINTR );
#else
    while( ret==-1L);
#endif

    return ret;
	

//      return TEMP_FAILURE_RETRY (::select (FD_SETSIZE, 
//  					 &set, NULL, NULL, 
//  					 _ptimeout));
    
}

int Sock::accept(int f(Sock *sock, void *arg), void *arg, bool use_sockthread) {

    int nsock;
    struct sockaddr_in clnt_adr;
    int clnt_len = sizeof(clnt_adr);

    if ((nsock = ::accept(sock, (struct sockaddr *)&clnt_adr, (socklen_t *)&clnt_len)) < 0)
	return 1;

    Sock *new_sock = new Sock(nsock);
    if(!use_sockthread){
	f(new_sock, arg);
	if(new_sock!=NULL)
	    delete new_sock;
    }
    else{
	fthread_arg *tharg = new fthread_arg(new_sock, arg, f);
#ifdef USE_PTHREAD
	pthread_attr_t attr;
	
	/* Initialize and set thread detached attribute */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	pthread_t thread_id;
	pthread_create(&thread_id, &attr, fthread, tharg);
	pthread_attr_destroy(&attr);
#else
	unsigned long thread_id = _beginthread( fthread, 0, tharg);
#endif
    }
    return 0;
}

int
Sock::getpeername(string& hostname){
    socklen_t len;
    struct sockaddr_in sin;
    struct hostent *host;


    len = sizeof sin;
    if (::getpeername(sock, (struct sockaddr *) &sin, &len) < 0)
        return 1;
    else if ((host = gethostbyaddr((char *) &sin.sin_addr,
				   sizeof sin.sin_addr,
				   AF_INET)) == NULL)
	return 2;
    else {
	hostname = string(host->h_name);
	while (*host->h_addr_list) { 
	    struct in_addr in;
	    memcpy(&in.s_addr, *host->h_addr_list, sizeof(in.s_addr));
	    hostname += '\t'; hostname += inet_ntoa(in);
	    ++host->h_addr_list;
	}
    }
	//	hostname = string(host->h_name);
    
    return 0;
}

int TestEndian() {
    char b[2] = {0x00, 0x0F};
    short  *p = (short *) b;

    if (*p == 0x0F00) return LittleEndian;
    if (*p == 0x000F) return BigEndian;

    return 0;
}

int fswap(void *_buffer, size_t size, size_t count)  {
    unsigned int i,j;
    char tmp;
    char *h, *l, *buffer = (char *) _buffer;

    if (size <= 1)
	return 0;

    for (i = 0; i < count ; ++i) {
	l = buffer;
	h = l + size -1;
	for (j = 0; j < size/2; ++j) {
	    tmp = *l;
	    *l++ = *h;
	    *h-- = tmp;
	}
	buffer += size;
    }
    return 0;
}
