/**
 * Copyright (C) 2013 Quan Wang <wangq10@rpi.edu>,
 * Signal Analysis and Machine Perception Laboratory,
 * Department of Electrical, Computer, and Systems Engineering,
 * Rensselaer Polytechnic Institute, Troy, NY 12180, USA
 */

/** 
 * This is the C/MEX code of dynamic time warping of two signals
 *
 * compile: 
 *     mex dtw_c.c
 *
 * usage:
 *     d=dtw_c(s,t)  or  d=dtw_c(s,t,w)
 *     where s is signal 1, t is signal 2, w is window parameter 
 */


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define DIAG 0
#define HALF 1
#define DOUB 2

double vectorDistance(double *s, double *t, int ns, int nt, int k, int i, int j)
{
    double result=0;
    double ss,tt;
    int x;
    for(x=0;x<k;x++)
    {
        ss=s[i+ns*x];
        tt=t[j+nt*x];
        
        if((ss!=tt) && (ss==100 || tt==100)){
            result+=100;
           
        } else{
            result+=((ss-tt)*(ss-tt));
            
        }
    }
    result=sqrt(result);
   // if (result>12){
   //     result=12;
   // }
    return result;
}

double dtw_c(double *s, double *t, int w, int ns, int nt, int k)
{
    double d=0;
    int sizediff=ns-nt>0 ? ns-nt : nt-ns;
    double ** D;
    int ** backtracking;
    int i,j;
    int j1,j2;
    double cost,temp,min;
    int *x, *y;
    int pathlength = 0, ix, iy;
    FILE *fp, *fm, *fb;
    
    if(ns>=1.8*nt){
        return 9999;
    }
    
    x = (int *) malloc((ns+nt)*sizeof(int));
    y = (int *) malloc((ns+nt)*sizeof(int));
    
            
    // printf("ns=%d, nt=%d, w=%d, s[0]=%f, t[0]=%f\n",ns,nt,w,s[0],t[0]);
    
    
    if(w!=-1 && w<sizediff) w=sizediff; // adapt window size
    
    // create D
    D=(double **)malloc((ns+1)*sizeof(double *));
    backtracking=(int **)malloc((ns+1)*sizeof(int *));
    
    for(i=0;i<ns+1;i++)
    {
        D[i]=(double *)malloc((nt+1)*sizeof(double));
        backtracking[i]=(int *)malloc((nt+1)*sizeof(int));
    }
    
    // initialization
    for(j=0;j<nt+1;j++)
        {
            D[0][j]=0;
            D[1][j]=0;
            backtracking[0][j]=-1;
            backtracking[1][j]=-1;
        }
    
       
    for(i=2;i<ns+1;i++)
    {
       /* for(j=0;j<2;j++)
        {
            D[i][j]=0;
            backtracking[i][j]=-1;
        }
          */
        for(j=0;j<nt+1;j++){
            D[i][j]=-1;
            backtracking[i][j]=-1;
        }
    }
    //D[0][0]=0;
    
    // dynamic programming
    //printf("Dynamic programming\n");
    for(i=2;i<=ns;i++)
    {
        
        if(w==-1)
        {
            j1=2;
            j2=nt;
            //printf("Dins IF sospitos\n");
        }
        else
        {
            j1= i-w>1 ? i-w : 2;
            j2= i+w<nt ? i+w : nt;
            //printf("Dins ELSE ok\n");
        }
        for(j=j1;j<=j2;j++)
        {
            cost=vectorDistance(s,t,ns,nt,k,i-1,j-1);
            //printf("Cost = %f\n", cost);
            
            temp=D[i-1][j-2];
            backtracking[i][j]=HALF;
            //printf("Cas HALF\n");
            if(D[i-2][j-1]!=-1) 
            {
                if(temp==-1 || D[i-2][j-1]<temp) {
                    temp=D[i-2][j-1];
                    backtracking[i][j]=DOUB;
                    //printf("Cas DOUB\n");
                }
            }
                        
            if(temp!=-1){
                D[i][j]=2*cost+temp;
            }
            
            if(D[i-1][j-1]!=-1) 
            {
                if(temp==-1 || D[i-1][j-1]+cost < D[i][j]) {
                    temp=D[i-1][j-1];
                    backtracking[i][j]=DIAG;
                    //printf("Cas DIAG\n");
                    D[i][j]=cost+temp;
                }
            }
            if(j==2 && i==2){
                
            	//printf("Temp = %f\t Cost = %f\t D[i][j] = %f\n", temp, cost, D[i][j]); 
            }
            //D[i][j]=cost+temp;
        }
    }
    
//      fm = fopen("matrix.txt","wt");
//      fb = fopen("backtracking.txt","wt");
//     
//      fprintf(fm,"-2\t-2\t");
//      for(i=1;i<nt;i++){
//          fprintf(fm,"%f\t",t[i]);
//      }
//      fprintf(fm,"\n");
//    
//      
//      for(i=1;i<=ns;i++){
//          fprintf(fm,"%f\t",s[i]);
//          for(j=1;j<=nt;j++){
//              fprintf(fm,"%f\t",D[i][j]);
//              fprintf(fb,"%d\t",backtracking[i][j]);
//          }
//       fprintf(fm,"\n");
//       fprintf(fb,"\n");
//     }
//     
//     fclose(fb);
//     fclose(fm);
    
    
    //printf("Min search\n");
    
    min = D[ns][2];
    ix = ns;
    iy = 2;
   // printf("Min = %f\t D[ns][0] = %f\t D[ns][1] = %f\n",min,D[ns][0],D[ns][1]);
    for(j=2;j<nt;j++){
        //printf("D[ns][j] = %d\n Min = %d", D[ns][j], min);
        if(D[ns][j]<=min || min==-1){
            min=D[ns][j];  
            iy = j;
            //printf("Dins de IF\n");
        }
    //printf("For it %d iy = %d\n", j-1, iy);   
    }
    //printf("IY final %d\n IX final %d\n NS %d NT %d MIN %f\n",iy,ix,ns,nt,min);
    if(min!=-1){
        d=min;
    }else d=3.141592;
    
    
    //Path
    //printf("Path\n");
    x[pathlength]=ix;
    y[pathlength]=iy;
    pathlength++;
    
    while(ix>1){
        if(backtracking[ix][iy]==DIAG){
            ix=ix-1;
            iy=iy-1;
            //printf("Backtracking DIAG ix = %d iy = %d\n",ix,iy);  
            
        } else if(backtracking[ix][iy]==DOUB){
            iy=iy-1;
            ix=ix-2;
            //printf("Backtracking DOUB ix = %d iy = %d\n",ix,iy);
        } else if(backtracking[ix][iy]==HALF){
            ix=ix-1;
            iy=iy-2;
            //printf("Backtracking HALF ix = %d iy = %d\n",ix,iy);
        } else{
            printf("Error a Backtracking, ns = %d nt = %d\n", ns,nt);
            ix=0;      
        }
        
         x[pathlength]=ix;
         y[pathlength]=iy;
         pathlength++;  
       
      }
    
   // printf("Pathlength = %d\n", pathlength);
    //printf("Fopen - While ended\n");
    
    fp = fopen("path.txt","wt");
    
    for(i=pathlength-1;i>=0;i--){
       // printf("Writing path i = %d, with x[i] = %d, y[i] = %d, D[x[i]][y[i]] = %f\n",i,x[i],y[i],D[x[i]][y[i]]);
        //printf("For fopen it %d\n",i);
        //printf("%d\t%d\t%f\n",x[i],y[i],D[x[i]][y[i]]);
        fprintf(fp,"%d\t%d\t%f\n",x[i],y[i],D[x[i]][y[i]]);
      
    }
    fclose(fp);
//     
    
    
   // d=D[ns][nt]; //BUSCAR MNIM?
    
    /* view matrix D */
    /*
    for(i=0;i<ns+1;i++)
    {
        for(j=0;j<nt+1;j++)
        {
            printf("%f  ",D[i][j]);
        }
        printf("\n");
    }
    */ 
    
    // free D
    for(i=0;i<ns+1;i++)
    {
        free(D[i]);
        free(backtracking[i]);
    }
    free(D);
    free(backtracking);
    

    return d/pathlength;
}


