/*-------------------------------------------------------------------*
|GAIA - Software platform to integrate Speech Translation Technologies
|Funded by the European Union (LC-STAR project, IST-2001-32216) and
|the Spanish Government (ALIADO project, TIC2002-04447-C02).
|
|Copyright (C) Javier Perez		javier.perez@tsc.upc.edu
|	       Antonio Bonafonte	antonio.bonafonte@upc.edu
|	       Universitat Politecnica de Catalunya, Barcelona, Spain.
|
|Gaia (main program and related libraries and utilities) is free
|software; you can redistribute it and/or modify it under the terms of
|the GNU General Public License as published by the Free Software
|Foundation; either version 2 of the License, or (at your option) any
|later version.
|
|This program is distributed in the hope that it will be useful,
|but WITHOUT ANY WARRANTY; without even the implied warranty of
|MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
|GNU General Public License for more details.
|
|You should have received a copy of the GNU General Public License
|along with this program; if not, write to the Free Software
|Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
|USA.
*--------------------------------------------------------------------*/
#ifndef SOCK_H
#define SOCK_H

#include <sys/types.h>
#include <string>
using namespace std;
#ifndef WIN32
#include <unistd.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifndef USE_PTHREAD
#define USE_PTHREAD
#endif
#else

#include <winsock2.h>

#define socklen_t int
#endif

class Sock {
  int sock;
  bool swap;

public:

  Sock();
  Sock(int sock);
#ifndef WIN32
  ~Sock() {::close(sock);}
#else
  ~Sock() {closesocket(sock);}
#endif
  int get_sock() {return sock;}
  int connect(string server, int port);
  int connect(const char *server, int port);
  int listen(int port);

  void close() {
#ifndef WIN32
      ::close(sock);
#else
      closesocket(sock);
#endif
  }

  long write(const void *buf, size_t size, size_t count) const;
  long read(void *buf, size_t size, size_t count) const;

  long read(void *buf, size_t count) const;
  long write(const void *buf, size_t count) const;

  long write(const string &data) const;
  long read(string &data, size_t count) const;

  int select(long timeout_usec = -1);
  int accept(int f(Sock *sock, void *arg), 
	     void *arg, bool use_sockthread=true);

  int getpeername(string& hostname);
};

#endif
